{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "fatty-appreciation",
   "metadata": {},
   "source": [
    "## MNIST Non IID "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "id": "alternative-football",
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch\n",
    "import torchvision\n",
    "\n",
    "import numpy as np\n",
    "import math"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "id": "capable-province",
   "metadata": {},
   "outputs": [],
   "source": [
    "mnist_trainset = torchvision.datasets.MNIST(root='./data', train=True, download=True, transform=torchvision.transforms.Compose([\n",
    "                               torchvision.transforms.ToTensor(),\n",
    "                               torchvision.transforms.Normalize(\n",
    "                                 (0.1307,), (0.3081,))\n",
    "                             ]))\n",
    "mnist_testset = torchvision.datasets.MNIST(root='./data', train=False, download=True, transform=torchvision.transforms.Compose([\n",
    "                               torchvision.transforms.ToTensor(),\n",
    "                               torchvision.transforms.Normalize(\n",
    "                                 (0.1307,), (0.3081,))\n",
    "                             ]))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "id": "criminal-immunology",
   "metadata": {},
   "outputs": [],
   "source": [
    "## Parameters:\n",
    "n_epochs = 3\n",
    "batch_size_train = 10000\n",
    "batch_size_test = 500\n",
    "log_interval = 500"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "id": "meaning-denver",
   "metadata": {},
   "outputs": [],
   "source": [
    "train_loader = torch.utils.data.DataLoader(mnist_trainset,batch_size=batch_size_train, shuffle=True)\n",
    "test_loader = torch.utils.data.DataLoader(mnist_testset,batch_size=batch_size_test, shuffle=False)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "id": "experienced-shoulder",
   "metadata": {},
   "outputs": [],
   "source": [
    "import papayaclient"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "id": "friendly-philosophy",
   "metadata": {},
   "outputs": [],
   "source": [
    "class TheModel(torch.nn.Module):\n",
    "\n",
    "    def __init__(self):\n",
    "        super(TheModel, self).__init__()\n",
    "\n",
    "        self.linear1 = torch.nn.Linear(784, 400)\n",
    "        self.linear2 = torch.nn.Linear(400, 10)\n",
    "        self.relu = torch.nn.ReLU()\n",
    "\n",
    "    def forward(self, x):\n",
    "        x1 = x.flatten(start_dim = 1)\n",
    "        return self.linear2(self.relu(self.linear1(x1)))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "id": "tough-welcome",
   "metadata": {},
   "outputs": [],
   "source": [
    "clients = []\n",
    "list_of_data = []\n",
    "list_of_labels = []\n",
    "for batchno, (ex_data, ex_labels) in enumerate(train_loader):\n",
    "    list_of_data.append(ex_data)\n",
    "    list_of_labels.append(ex_labels)\n",
    "data, labels = torch.cat(list_of_data), torch.cat(list_of_labels)\n",
    "np.random.seed(42)\n",
    "\n",
    "node_data = [[] for _ in range(6)]\n",
    "node_labels = [[] for _ in range(6)]\n",
    "for num in range(9):\n",
    "    data_by_class = data[labels == num]\n",
    "    label_by_class = labels[labels == num]\n",
    "\n",
    "    idx = [0] + sorted(np.random.choice(len(data_by_class)-1, 5, replace=False)+1) + [len(data_by_class)]\n",
    "    for i in range(6):\n",
    "        ex_data = data[idx[i]:idx[i+1]]\n",
    "        ex_labels = labels[idx[i]:idx[i+1]]\n",
    "        node_data[i].append(ex_data)\n",
    "        node_labels[i].append(ex_labels)\n",
    "for i in range(6):\n",
    "    ex_data = torch.cat(node_data[i])\n",
    "    ex_labels = torch.cat(node_labels[i])\n",
    "    rand_idx = torch.randperm(len(ex_data))\n",
    "    ex_data = ex_data[rand_idx]\n",
    "    ex_labels = ex_labels[rand_idx]\n",
    "    clients.append(papayaclient.PapayaClient(dat = ex_data,\n",
    "                                            labs = ex_labels,\n",
    "                                            batch_sz = 500,\n",
    "                                            num_partners = 5,\n",
    "                                            model_class = TheModel,\n",
    "                                            loss_fn = torch.nn.CrossEntropyLoss))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "id": "brilliant-great",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "1\n",
      "2\n",
      "3\n",
      "4\n",
      "5\n",
      "1\n",
      "2\n",
      "3\n",
      "4\n",
      "5\n",
      "1\n",
      "2\n",
      "3\n",
      "4\n",
      "5\n",
      "1\n",
      "2\n",
      "3\n",
      "4\n",
      "5\n",
      "1\n",
      "2\n",
      "3\n",
      "4\n",
      "5\n",
      "1\n",
      "2\n",
      "3\n",
      "4\n",
      "5\n",
      "6\n",
      "7\n",
      "8\n",
      "9\n",
      "10\n",
      "6\n",
      "7\n",
      "8\n",
      "9\n",
      "10\n",
      "6\n",
      "7\n",
      "8\n",
      "9\n",
      "10\n",
      "6\n",
      "7\n",
      "8\n",
      "9\n",
      "10\n",
      "6\n",
      "7\n",
      "8\n",
      "9\n",
      "10\n",
      "6\n",
      "7\n",
      "8\n",
      "9\n",
      "10\n",
      "11\n",
      "12\n",
      "13\n",
      "14\n",
      "15\n",
      "11\n",
      "12\n",
      "13\n",
      "14\n",
      "15\n",
      "11\n",
      "12\n",
      "13\n",
      "14\n",
      "15\n",
      "11\n",
      "12\n",
      "13\n",
      "14\n",
      "15\n",
      "11\n",
      "12\n",
      "13\n",
      "14\n",
      "15\n",
      "11\n",
      "12\n",
      "13\n",
      "14\n",
      "15\n",
      "16\n",
      "17\n",
      "18\n",
      "19\n",
      "20\n",
      "16\n",
      "17\n",
      "18\n",
      "19\n",
      "20\n",
      "16\n",
      "17\n",
      "18\n",
      "19\n",
      "20\n",
      "16\n",
      "17\n",
      "18\n",
      "19\n",
      "20\n",
      "16\n",
      "17\n",
      "18\n",
      "19\n",
      "20\n",
      "16\n",
      "17\n",
      "18\n",
      "19\n",
      "20\n",
      "21\n",
      "22\n",
      "23\n",
      "24\n",
      "25\n",
      "21\n",
      "22\n",
      "23\n",
      "24\n",
      "25\n",
      "21\n",
      "22\n",
      "23\n",
      "24\n",
      "25\n",
      "21\n",
      "22\n",
      "23\n",
      "24\n",
      "25\n",
      "21\n",
      "22\n",
      "23\n",
      "24\n",
      "25\n",
      "21\n",
      "22\n",
      "23\n",
      "24\n",
      "25\n",
      "26\n",
      "27\n",
      "28\n",
      "29\n",
      "30\n",
      "26\n",
      "27\n",
      "28\n",
      "29\n",
      "30\n",
      "26\n",
      "27\n",
      "28\n",
      "29\n",
      "30\n",
      "26\n",
      "27\n",
      "28\n",
      "29\n",
      "30\n",
      "26\n",
      "27\n",
      "28\n",
      "29\n",
      "30\n",
      "26\n",
      "27\n",
      "28\n",
      "29\n",
      "30\n",
      "31\n",
      "32\n",
      "33\n",
      "34\n",
      "35\n",
      "31\n",
      "32\n",
      "33\n",
      "34\n",
      "35\n",
      "31\n",
      "32\n",
      "33\n",
      "34\n",
      "35\n",
      "31\n",
      "32\n",
      "33\n",
      "34\n",
      "35\n",
      "31\n",
      "32\n",
      "33\n",
      "34\n",
      "35\n",
      "31\n",
      "32\n",
      "33\n",
      "34\n",
      "35\n",
      "36\n",
      "37\n",
      "38\n",
      "39\n",
      "40\n",
      "36\n",
      "37\n",
      "38\n",
      "39\n",
      "40\n",
      "36\n",
      "37\n",
      "38\n",
      "39\n",
      "40\n",
      "36\n",
      "37\n",
      "38\n",
      "39\n",
      "40\n",
      "36\n",
      "37\n",
      "38\n",
      "39\n",
      "40\n",
      "36\n",
      "37\n",
      "38\n",
      "39\n",
      "40\n",
      "41\n",
      "42\n",
      "43\n",
      "44\n",
      "45\n",
      "41\n",
      "42\n",
      "43\n",
      "44\n",
      "45\n",
      "41\n",
      "42\n",
      "43\n",
      "44\n",
      "45\n",
      "41\n",
      "42\n",
      "43\n",
      "44\n",
      "45\n",
      "41\n",
      "42\n",
      "43\n",
      "44\n",
      "45\n",
      "41\n",
      "42\n",
      "43\n",
      "44\n",
      "45\n",
      "46\n",
      "47\n",
      "48\n",
      "49\n",
      "50\n",
      "46\n",
      "47\n",
      "48\n",
      "49\n",
      "50\n",
      "46\n",
      "47\n",
      "48\n",
      "49\n",
      "50\n",
      "46\n",
      "47\n",
      "48\n",
      "49\n",
      "50\n",
      "46\n",
      "47\n",
      "48\n",
      "49\n",
      "50\n",
      "46\n",
      "47\n",
      "48\n",
      "49\n",
      "50\n",
      "51\n",
      "52\n",
      "53\n",
      "54\n",
      "55\n",
      "51\n",
      "52\n",
      "53\n",
      "54\n",
      "55\n",
      "51\n",
      "52\n",
      "53\n",
      "54\n",
      "55\n",
      "51\n",
      "52\n",
      "53\n",
      "54\n",
      "55\n",
      "51\n",
      "52\n",
      "53\n",
      "54\n",
      "55\n",
      "51\n",
      "52\n",
      "53\n",
      "54\n",
      "55\n",
      "56\n",
      "57\n",
      "58\n",
      "59\n",
      "60\n",
      "56\n",
      "57\n",
      "58\n",
      "59\n",
      "60\n",
      "56\n",
      "57\n",
      "58\n",
      "59\n",
      "60\n",
      "56\n",
      "57\n",
      "58\n",
      "59\n",
      "60\n",
      "56\n",
      "57\n",
      "58\n",
      "59\n",
      "60\n",
      "56\n",
      "57\n",
      "58\n",
      "59\n",
      "60\n",
      "61\n",
      "62\n",
      "63\n",
      "64\n",
      "65\n",
      "61\n",
      "62\n",
      "63\n",
      "64\n",
      "65\n",
      "61\n",
      "62\n",
      "63\n",
      "64\n",
      "65\n",
      "61\n",
      "62\n",
      "63\n",
      "64\n",
      "65\n",
      "61\n",
      "62\n",
      "63\n",
      "64\n",
      "65\n",
      "61\n",
      "62\n",
      "63\n",
      "64\n",
      "65\n",
      "66\n",
      "67\n",
      "68\n",
      "69\n",
      "70\n",
      "66\n",
      "67\n",
      "68\n",
      "69\n",
      "70\n",
      "66\n",
      "67\n",
      "68\n",
      "69\n",
      "70\n",
      "66\n",
      "67\n",
      "68\n",
      "69\n",
      "70\n",
      "66\n",
      "67\n",
      "68\n",
      "69\n",
      "70\n",
      "66\n",
      "67\n",
      "68\n",
      "69\n",
      "70\n",
      "71\n",
      "72\n",
      "73\n",
      "74\n",
      "75\n",
      "71\n",
      "72\n",
      "73\n",
      "74\n",
      "75\n",
      "71\n",
      "72\n",
      "73\n",
      "74\n",
      "75\n",
      "71\n",
      "72\n",
      "73\n",
      "74\n",
      "75\n",
      "71\n",
      "72\n",
      "73\n",
      "74\n",
      "75\n",
      "71\n",
      "72\n",
      "73\n",
      "74\n",
      "75\n",
      "76\n",
      "77\n",
      "78\n",
      "79\n",
      "80\n",
      "76\n",
      "77\n",
      "78\n",
      "79\n",
      "80\n",
      "76\n",
      "77\n",
      "78\n",
      "79\n",
      "80\n",
      "76\n",
      "77\n",
      "78\n",
      "79\n",
      "80\n",
      "76\n",
      "77\n",
      "78\n",
      "79\n",
      "80\n",
      "76\n",
      "77\n",
      "78\n",
      "79\n",
      "80\n",
      "81\n",
      "82\n",
      "83\n",
      "84\n",
      "85\n",
      "81\n",
      "82\n",
      "83\n",
      "84\n",
      "85\n",
      "81\n",
      "82\n",
      "83\n",
      "84\n",
      "85\n",
      "81\n",
      "82\n",
      "83\n",
      "84\n",
      "85\n",
      "81\n",
      "82\n",
      "83\n",
      "84\n",
      "85\n",
      "81\n",
      "82\n",
      "83\n",
      "84\n",
      "85\n",
      "86\n",
      "87\n",
      "88\n",
      "89\n",
      "90\n",
      "86\n",
      "87\n",
      "88\n",
      "89\n",
      "90\n",
      "86\n",
      "87\n",
      "88\n",
      "89\n",
      "90\n",
      "86\n",
      "87\n",
      "88\n",
      "89\n",
      "90\n",
      "86\n",
      "87\n",
      "88\n",
      "89\n",
      "90\n",
      "86\n",
      "87\n",
      "88\n",
      "89\n",
      "90\n",
      "91\n",
      "92\n",
      "93\n",
      "94\n",
      "95\n",
      "91\n",
      "92\n",
      "93\n",
      "94\n",
      "95\n",
      "91\n",
      "92\n",
      "93\n",
      "94\n",
      "95\n",
      "91\n",
      "92\n",
      "93\n",
      "94\n",
      "95\n",
      "91\n",
      "92\n",
      "93\n",
      "94\n",
      "95\n",
      "91\n",
      "92\n",
      "93\n",
      "94\n",
      "95\n",
      "96\n",
      "97\n",
      "98\n",
      "99\n",
      "100\n",
      "96\n",
      "97\n",
      "98\n",
      "99\n",
      "100\n",
      "96\n",
      "97\n",
      "98\n",
      "99\n",
      "100\n",
      "96\n",
      "97\n",
      "98\n",
      "99\n",
      "100\n",
      "96\n",
      "97\n",
      "98\n",
      "99\n",
      "100\n",
      "96\n",
      "97\n",
      "98\n",
      "99\n",
      "100\n"
     ]
    }
   ],
   "source": [
    "## Train the Nodes\n",
    "num_epochs_total = 100\n",
    "num_epochs_per_swap = 5\n",
    "num_times = (num_epochs_total // num_epochs_per_swap)\n",
    "for i in range(0, num_times):\n",
    "    for n in clients:\n",
    "        for j in range(0, num_epochs_per_swap):\n",
    "            n.model_train_epoch()\n",
    "            print(n.epochs_trained)\n",
    "    if i > 1 and i < num_times - 1 :\n",
    "        for n in clients:\n",
    "            n.select_partners(3)\n",
    "        for n in clients:\n",
    "            for i in range(0, 4) :\n",
    "                n.update_partner_weights()\n",
    "            n.average_partners()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "id": "periodic-recall",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "node3811epoch 99 loss 0.2988002896308899\n",
      "node2244epoch 99 loss 0.30985724925994873\n",
      "node1856epoch 99 loss 0.12980203330516815\n",
      "node4116epoch 99 loss 0.19149348139762878\n",
      "node4327epoch 99 loss 0.24858054518699646\n",
      "node689epoch 99 loss 0.19274069368839264\n"
     ]
    }
   ],
   "source": [
    "for c in clients :\n",
    "    print(c.logs['stringy'][99])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "id": "marine-kidney",
   "metadata": {},
   "outputs": [],
   "source": [
    "accuracies = {}\n",
    "with torch.no_grad():\n",
    "    for i in clients :\n",
    "        accuracies_node = []\n",
    "        for batchno, (ex_data, ex_labels) in enumerate(test_loader) :\n",
    "            accuracies_node.append(((i.model.forward(ex_data).argmax(dim = 1) == ex_labels).float().mean()).item())\n",
    "        accuracies[i.node_id] = np.array(accuracies_node).mean()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "id": "comparable-bicycle",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "{3811: 0.9123000055551529,\n",
       " 2244: 0.9122999995946884,\n",
       " 1856: 0.9106000006198883,\n",
       " 4116: 0.9099000036716461,\n",
       " 4327: 0.9087999999523163,\n",
       " 689: 0.9097000092267991}"
      ]
     },
     "execution_count": 11,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "accuracies"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "surprising-latex",
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.8"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
